/****************************************************************************/
/*  Copyright 2005, Trustees of Indiana University                          */
/*  Author: Sidharth Thakur                                                 */
/*  Last modified: 14 January 2006                                          */
/*  License available at: <belt-trick-* folder>/docs/downloads/License.txt  */
/****************************************************************************/
///////////////////////////////////////////////////////////////////////////////
//
// $Implemenatation:  QuaternionSet.cpp$
//
// NAME: MyOGLWidget
//
// BASE  CLASSES: none
//  
// PURPOSE: Implemenatation for class QuaternionSet
//
// AUTHOR: Sidharth Thakur
//
///////////////////////////////////////////////////////////////////////////////
#include "QuaternionSet.h"

static GLUquadricObj *qobjCYL, *qobjSPH;
static const GLfloat no_mat5[]					= {0.0f,0.0f,0.0f,1.0f};
static const GLfloat mat_ambient5[]				= {0.6f,0.4f,0.0f,1.0f}; 
static const GLfloat mat_ambient_color5[]		= {0.5f,0.5f,0.5f,1.0f};
static const GLfloat mat_diffuse5[]				= {0.85f,0.45f,0.1f,1.0f};
static const GLfloat mat_specular5[]			= {1.0f,1.0f,1.0f,1.0f};
static const GLfloat no_shininess5[]			= {0.0f};
static const GLfloat low_shininess5[]			= {10.0f};
static const GLfloat high_shininess5[]			= {100.0f};
static const GLfloat mat_emission5[]			= {0.3f,0.2f,0.2f,0.0f};

QuaternionSet::QuaternionSet() {

	id = 0;
	initObject();
	initStructs();
}

QuaternionSet::~QuaternionSet() 
{
	int i = 0;
	printf("in ~QuaternionSet()\n");
	
	//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	//Note: Can't do this as the objects were
	//		decleared but not defined (msdn)
	//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	//zap(qobjCYL);		
	//zap(qobjSPH);	
	
    for(i=0;i<MAX_POINTS;i++) 
        free(quatPointSet[i]);
	my_free(quatPointSet);	

	for(i=0;i<MAX_POINTS;i++) 
        free(quatPointSetNew[i]);
	my_free(quatPointSetNew);

	my_free(pset);
	my_free(tset);
	my_free(tbset);
}

void QuaternionSet::initObject ( ) 
{

}

void QuaternionSet::initStructs ( ) 
{
	int	i = 0;

	qobjCYL		= gluNewQuadric();
	qobjSPH		= gluNewQuadric();

	////////////////////////////////////
	// list of quaternions interpolated
	////////////////////////////////////
	quatPointSet		= (quaternion**) my_malloc(sizeof(quaternion)*MAX_POINTS);
    for(i=0;i<MAX_POINTS;i++) {
        quatPointSet[i] = (quaternion*) my_malloc(sizeof(quaternion));
    }
    for(i=0;i<MAX_POINTS;i++) {
        
        quatPointSet[i]->r = 0.0;
        quatPointSet[i]->i = 0.0;
        quatPointSet[i]->j = 0.0;
        quatPointSet[i]->k = 0.0;
    }	

	quatPointSetNew		= (quaternion**) my_malloc(sizeof(quaternion)*MAX_POINTS);
    for(i=0;i<MAX_POINTS;i++) {
        quatPointSetNew[i] = (quaternion*) my_malloc(sizeof(quaternion));
    }
    for(i=0;i<MAX_POINTS;i++) {
        
        quatPointSetNew[i]->r = 0.0;
        quatPointSetNew[i]->i = 0.0;
        quatPointSetNew[i]->j = 0.0;
        quatPointSetNew[i]->k = 0.0;
    }	
	
	pset = (Vertex*) my_malloc(sizeof(Vertex)*MAX_POINTS); 
	tset = (Vertex*) my_malloc(sizeof(Vertex)*MAX_POINTS); 
	tbset= (Vertex*) my_malloc(sizeof(Vertex)*MAX_POINTS); 

}

void QuaternionSet::setPointSet ( Vertex *refPointSet, int numPoints ) 
{
	int i = numPoints;
	
	for(i=0; i<numPoints; i++) {
		pset[i].p4.d[0] = refPointSet[i].p3.d[0];
		pset[i].p4.d[1] = refPointSet[i].p3.d[1];
		pset[i].p4.d[2] = refPointSet[i].p3.d[2];
	}
	numpset = numPoints;
}

void QuaternionSet::setTangSet ( Vertex *refTangSet, int numPoints ) 
{
	int i = numPoints;
	
	for(i=0; i<numPoints; i++) {
		tset[i].p4.d[0] = refTangSet[i].p3.d[0];
		tset[i].p4.d[1] = refTangSet[i].p3.d[1];
		tset[i].p4.d[2] = refTangSet[i].p3.d[2];
	}
	numtset = numPoints;
}

void QuaternionSet::setTubePointSet ( Vertex *pointList, int numPoints )
{
	int		i = 0;
	Point3	vec;
	Vertex* l_pointList = pointList;
	l_pointList = NULL;
	printf("-------\n");
	for(i=0; i<numPoints; i++) {

		switch(whichQuatComponent) {

			case SHOWXYZ:
				vec.d[0] = quatPointSet[i]->i;
				vec.d[1] = quatPointSet[i]->j;
				vec.d[2] = quatPointSet[i]->k;
				break;

			case SHOWWYZ:
				vec.d[0] = quatPointSet[i]->r;
				vec.d[1] = quatPointSet[i]->j;
				vec.d[2] = quatPointSet[i]->k;
				break;
			
			case SHOWWXY:
				vec.d[0] = quatPointSet[i]->r;
				vec.d[1] = quatPointSet[i]->i;
				vec.d[2] = quatPointSet[i]->j;
				break;

			case SHOWWXZ:
				vec.d[0] = quatPointSet[i]->r;
				vec.d[1] = quatPointSet[i]->i;
				vec.d[2] = quatPointSet[i]->k;
				break;

			default:
				break;
		}

		tbset[i].p3.d[0] = vec.d[0];
		tbset[i].p3.d[1] = vec.d[1];
		tbset[i].p3.d[2] = vec.d[2];
		//printf("[%d] tb = %f, %f, %f\n",i,tbset[i].p3.d[0],tbset[i].p3.d[1],tbset[i].p3.d[2]);
	}
}

void QuaternionSet::computeQuaternionSet ( int mode )
{
	int			i = 0;
	double		theta = 0.0;
	double		v[3];

	if(mode==1){
		for(i=0; i<numtset; i++) {
			v[0]=tset[i].p4.d[0];
			v[1]=tset[i].p4.d[1];
			v[2]=tset[i].p4.d[2];
			//printf("v = %f,%f,%f\n",v[0],v[1],v[2]);
			makeFamilyOfQuaternions(theta,v,quatPointSet[i]);
			quatPointSetNew[i]->r = quatPointSet[i]->r;
			quatPointSetNew[i]->i = quatPointSet[i]->i;
			quatPointSetNew[i]->j = quatPointSet[i]->j;
			quatPointSetNew[i]->k = quatPointSet[i]->k;
		}
		numqset = numtset;
	}
	else if(mode==2) {
		for(i=0; i<numqset; i++){
			quatPointSet[i]->r = quatPointSetNew[i]->r;
			quatPointSet[i]->i = quatPointSetNew[i]->i;
			quatPointSet[i]->j = quatPointSetNew[i]->j;
			quatPointSet[i]->k = quatPointSetNew[i]->k;
		}
	}
}

void QuaternionSet::generateQuaternionRings ( void ) {
	int			i = 0;
	double		theta = 0.0;
	double		v[3];
	Point3		vec;
	quaternion* q;
	q = (quaternion*) my_malloc(sizeof(quaternion));

	glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT);
	for(i=0;i<numqset;i++) {
		theta = 0.0;
		glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);

		glBegin(GL_LINE_STRIP);
		while(theta<=4*M_PI) {
			v[0]=tset[i].p4.d[0];
			v[1]=tset[i].p4.d[1];
			v[2]=tset[i].p4.d[2];
			makeFamilyOfQuaternions(theta,v,q);
			theta = theta + 0.5;

			switch(whichQuatComponent) {

				case SHOWXYZ:
					vec.d[0] = q->i;
					vec.d[1] = q->j;
					vec.d[2] = q->k;
					break;

				case SHOWWYZ:
					vec.d[0] = q->r;
					vec.d[1] = q->j;
					vec.d[2] = q->k;
					break;
				
				case SHOWWXY:
					vec.d[0] = q->r;
					vec.d[1] = q->i;
					vec.d[2] = q->j;
					break;

				case SHOWWXZ:
					vec.d[0] = q->r;
					vec.d[1] = q->i;
					vec.d[2] = q->k;
					break;

				default:
					break;
			}
			glColor3f(0.0f,0.0f,0.0f);
			/*glPushMatrix();
				glTranslatef(vec.d[0],vec.d[1],vec.d[2]);
				glutSolidSphere(0.0075,8,8);
			glPopMatrix();
			*/
			glVertex3f(vec.d[0],vec.d[1],vec.d[2]);
		}
		glEnd();
		glPopAttrib();
	}
	my_free(q);
}

void QuaternionSet::makeFamilyOfQuaternions	( double theta, double *v, quaternion* qf ) 
{
	double		z[] = {0,0,1}; 
	double		crossZV[3];
	double		len;
	double		phi = 0.3;
	//double		mat[16];
	//double		t[3],r[3],s[3];
	quaternion	*qq,*qb;

	qq	 = (quaternion*) my_malloc(sizeof(quaternion));
	qb	 = (quaternion*) my_malloc(sizeof(quaternion));

	//////////////////////////////////////////////////////////////////

	////////
	// qq //
	////////
	//printf("theta = %f\n",theta);
	quatBuildQuaternion(theta,v,qq);
	//printf("qq = %f, %f, %f %f\n",qq->r,qq->i,qq->j,qq->k);

	////////
	// qb //
	////////
	cross3D(z,v,crossZV);
	phi = acos(dot3D(z,v));
	len = length3D(crossZV);
	crossZV[0] /= len;
	crossZV[1] /= len;
	crossZV[2] /= len;
	quatBuildQuaternion(phi,v,qb);
	//printf("qb = %f, %f, %f %f\n",qb->r,qb->i,qb->j,qb->k);

	////////
	// qf //
	////////
	qf->r = qq->r*qb->r - qq->i*qb->i - qq->j*qb->j - qq->k*qb->k;
	qf->i = qq->i*qb->r + qq->r*qb->i + qq->j*qb->k - qq->k*qb->j;
	qf->j = qq->j*qb->r + qq->r*qb->j + qq->k*qb->i - qq->i*qb->k;
	qf->k = qq->k*qb->r + qq->r*qb->k + qq->i*qb->j - qq->j*qb->i;
	//quatConjugate(qf);
	//printf("qf = %f, %f, %f %f\n",qf->r,qf->i,qf->j,qf->k);
	//////////////////////////////////////////////////////////////////
	my_free(qq);
	my_free(qb);

}

void QuaternionSet::optimize ( )
{
	//int			cnt = 0;
	int			i = 0;
	quaternion	Q1;
	quaternion	q1;

	double		origdist;
	double		initdist;
	double		delta=0.0;
	double		finaltheta;
	double		v[3];	//tangent vector

	double		theta0 = 0.0;
	//double		theta  = 0.0;
	double		h = 0.0;
	double		F = 0.0;
	//double		dF= 0.0;
	//double		mat[16];

	quaternion* q;
	quaternion* qfph;
	quaternion* qfmh;
	q = (quaternion*) my_malloc(sizeof(quaternion));
	qfph = (quaternion*) my_malloc(sizeof(quaternion));
	qfmh = (quaternion*) my_malloc(sizeof(quaternion));

	//printf("==========\n");
	i=0;
	quatPointSetNew[i]->r = quatPointSet[i]->r; 
	quatPointSetNew[i]->i = quatPointSet[i]->i;
	quatPointSetNew[i]->j = quatPointSet[i]->j;
	quatPointSetNew[i]->k = quatPointSet[i]->k;

	printf("INITIALLY > ");	curveQuatLength();

	for(i=0; i<(numqset-1); i++) {

		//printf("i = %f ---->\n",i);
		Q1.r = quatPointSet[i]->r;
		Q1.i = quatPointSet[i]->i;
		Q1.j = quatPointSet[i]->j;
		Q1.k = quatPointSet[i]->k;
		//printf("Q1 = %f, %f, %f, %f\n",Q1.r,Q1.i,Q1.j,Q1.k);
		
		h = 0.0001;
		if(numqset>50)
			h = 0.0005;
		if(numqset>100)
			h = 0.001;
		if(numqset>150)
			h = 0.005;

		q1.r = quatPointSet[i+1]->r;
		q1.i = quatPointSet[i+1]->i;
		q1.j = quatPointSet[i+1]->j;
		q1.k = quatPointSet[i+1]->k;
		
		
		origdist = quatQuatDistance(&q1,&Q1);
		//printf("origdist = %f\n",origdist);
		//get tangent vector
		v[0] = tset[i+1].p4.d[0];
		v[1] = tset[i+1].p4.d[1];
		v[2] = tset[i+1].p4.d[2];
		
		theta0 = 0.0;
		//get current quat
		initdist = 1000.00;

		while(theta0<4*M_PI) {
			//printf("[%d]\n",cnt++);
			//printf("theta0 = %f\n",theta0);


			makeFamilyOfQuaternions(theta0,v,q); 
			q1.r = q->r;
			q1.i = q->i;
			q1.j = q->j;
			q1.k = q->k;

			//define distance func
			F = quatQuatDistance(&q1,&Q1);
			if(F<initdist) {
				initdist = F;
				finaltheta = theta0;
			}

			delta = F - initdist;
			if(delta >0.005)
				theta0 = theta0+100*h;
			else 
				theta0 = theta0+h;
			//printf("delta = %f\n",delta);
		}
		//printf("orig = %f\n",origdist);	
		//printf("dist = %f\n",initdist);
		//printf("theta = %f\n",finaltheta);
		makeFamilyOfQuaternions(finaltheta,v,q); 
		//printf("q1 = %f, %f, %f, %f\n",q->r,q->i,q->j,q->k);
		quatPointSet[i+1]->r = q->r;
		quatPointSet[i+1]->i = q->i;
		quatPointSet[i+1]->j = q->j;
		quatPointSet[i+1]->k = q->k;

		quatPointSetNew[i+1]->r = q->r;
		quatPointSetNew[i+1]->i = q->i;
		quatPointSetNew[i+1]->j = q->j;
		quatPointSetNew[i+1]->k = q->k;
	}
	printf("FINALLY > ");curveQuatLength();
	my_free(q);
	my_free(qfph);
	my_free(qfmh);
}

void QuaternionSet::optimize2 ( )
{
	int			cnt = 0;
	int			i = 0;
	quaternion	Q1;
	quaternion	q1;
	double		v[3];	//tangent vector

	double		theta0		= 0.0;
	double		theta		= 0.0;
	double		thetaFwd	= 0.0;
	double		thetaBwd	= 0.0;
	double		initdist	= 0.0;
	double		origdist	= 0.0;

	double		qdist		= 0.0;
	double		qdistFwd	= 0.0;
	double		qdistBwd	= 0.0;
	double		delta		= 0.0;
	
	double		h			= 0.0;
	
	double		F			= 0.0;
	double		dF			= 0.0;
	double		ddF			= 0.0;

	double		divfactor	= 0.0;
	//double		mat[16];

	double		limit		= 0.0;

	quaternion* q;
	quaternion* qfph;
	quaternion* qfmh;
	q = (quaternion*) my_malloc(sizeof(quaternion));
	qfph = (quaternion*) my_malloc(sizeof(quaternion));
	qfmh = (quaternion*) my_malloc(sizeof(quaternion));

/*
	1. obtain keyframe quaternions Q1 and q1
	2. pick an initial phi01 angle for q1; 
	   set interval h
	3. determine vector v1
	4. make family of quaternion about v1
	5. start sliding:
		a. f(phi10) = 2acos(q10.Q1)
		b. df(phi10) = f(phi10+h)-f(phi10-h) / 2h
		c. phi11 = phi10 - f(phi10)/df(phi10);
	6. compute q1 = quaternion(q11);
	7. print distance = 2acos(q11.Q1)
*/

	//printf("initially : ");
	//curveQuatLength();

	h = 4*M_PI/100000.0;
	printf(">>>>>>\n");
	for(i=0; i<(1); i++) {

		limit	= 0.0001;
		theta0	= limit;
		theta	= limit;
		delta	= 1000.00;
		initdist= 1000.00;
		cnt		= 0;
		qdist	= 1000.00;

		Q1.r = quatPointSet[i]->r;
		Q1.i = quatPointSet[i]->i;
		Q1.j = quatPointSet[i]->j;
		Q1.k = quatPointSet[i]->k;
		//printf("Q1 = %f, %f, %f %f\n",Q1.r,Q1.i,Q1.j,Q1.k);

		q1.r = quatPointSet[i+1]->r;
		q1.i = quatPointSet[i+1]->i;
		q1.j = quatPointSet[i+1]->j;
		q1.k = quatPointSet[i+1]->k;

		origdist = quatQuatDistance(&q1,&Q1);
		
		theta0 = 2*acos(q1.r);
		printf("theta0 = %f\n",theta0);
		//get tangent vector
		v[0] = tset[i+1].p4.d[0];
		v[1] = tset[i+1].p4.d[1];
		v[2] = tset[i+1].p4.d[2];
		
	//while((qdist-origdist)/100 >= .05) {
			while(delta>limit) {

				//printf("[%d] ",cnt++);

				//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
				//@ angles
				thetaFwd= theta0+h;
				thetaBwd= theta0-h;
				if(thetaBwd<0.0)
					thetaBwd = limit;
				//printf(" theta= %f, theta0+h = %f, theta0-h = %f\n",theta,thetaFwd,thetaBwd);
				
				//@ quaternion distances
				makeFamilyOfQuaternions(theta0,v,q);
				qdist	 = quatQuatDistance(q,&Q1);
				makeFamilyOfQuaternions(thetaFwd,v,qfph);
				makeFamilyOfQuaternions(thetaBwd,v,qfmh);
				qdistFwd = quatQuatDistance(qfph,&Q1);
				qdistBwd = quatQuatDistance(qfmh,&Q1);
				printf("qdist = %f,\nqdistFwd = %f,\nqdistBwd = %f\n",qdist,qdistFwd,qdistBwd);

				//@ function, derivatives
				F	= qdist;
				//dF	= (qdistFwd - qdistBwd)/(2.0*h);
				
				dF	= (qdistFwd - qdist)/(2.0*h);
				printf("dF = %f\n",dF);
				ddF	= (qdistFwd + qdistBwd - 2.0*qdist); //NOTE! Error term O(h^2) is ignored
				//printf("ddF.1 = %f\n",ddF);
				ddF = ddF/(h*h);
				printf("ddF.2 = %f\n",ddF);
				//divfactor = dF/ddF;
				divfactor = F/dF;
				printf(" F = %f,\n dF = %f,\n ddF = %f,\n divfactor = %f\n",F,dF,ddF,divfactor);
				
				//@ updated theta
				theta = theta0 - divfactor;
				printf("theta0 = %f, theta = %f\n",theta0,theta);

				makeFamilyOfQuaternions(theta,v,q);
				qdist = quatQuatDistance(q,&Q1);
				//printf("qdist = %f\n",qdist);
				delta = initdist - qdist;

				if(delta<limit) {
					//printf("theta0 = %f, theta = %f\n",theta0,theta);
					//printf("initdist = %f, qdist = %f, delta = %f\n",initdist,qdist,delta);
					break;
				}
				//
				theta0 = theta;
				initdist = qdist;

				//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

			}
			printf("final THETA= %f\n",theta0);
			makeFamilyOfQuaternions(theta0,v,q); 
			quatPointSet[i+1]->r = q->r;
			quatPointSet[i+1]->i = q->i;
			quatPointSet[i+1]->i = q->j;
			quatPointSet[i+1]->i = q->k;
			qdist = quatQuatDistance(q,&Q1);
			//printf("q1 = %f, %f, %f %f\n",q1.r,q1.i,q1.j,q1.k);
			printf("origdist = %f\nqdist = %f\n",origdist,qdist);
			//printf("---------\n");
		}
		printf("origdist = %f\nqdist = %f\n",origdist,qdist);
	//}

	//printf("finally: ");
	//curveQuatLength();
	my_free(q);
	my_free(qfph);
	my_free(qfmh);

}

void QuaternionSet::curveQuatLength ( void )
{

	int		i		= 0;
	double	qdist	= 0.0;
	double	totaldist = 0.0;

	for(i=0; i<(numqset-1); i++) {
		qdist = quatQuatDistance(quatPointSet[i],quatPointSet[i+1]);
		totaldist += qdist;
		//printf("qdist = %f\n",qdist);
	}
	printf(" %f\n",totaldist);
}

void QuaternionSet::printVec ( quaternion* q ) 
{
	double angle;
	double v[3];
	angle = 2*acos(q->r);
	v[0] = q->i/sin(angle/2.0);
	v[1] = q->j/sin(angle/2.0);
	v[2] = q->k/sin(angle/2.0);
	printf("v = %f,%f,%f\n",v[0],v[1],v[2]);
}
void QuaternionSet::displayQuaternions ( ) 
{
	int i = 0;
	Point3 vec;

    glPushAttrib( GL_ALL_ATTRIB_BITS );

	if(showQuatCurve) {
		glDisable(GL_LIGHTING);
		glColor3f(1.0f,0.0f,0.0f);
		glLineWidth(2.0f);
		glBegin(GL_LINE_STRIP);

		for(i=0;i<numqset;i++) {
			switch(whichQuatComponent) {

				case SHOWXYZ:
					vec.d[0] = quatPointSetNew[i]->i;
					vec.d[1] = quatPointSetNew[i]->j;
					vec.d[2] = quatPointSetNew[i]->k;
					break;

				case SHOWWYZ:
					vec.d[0] = quatPointSetNew[i]->r;
					vec.d[1] = quatPointSetNew[i]->j;
					vec.d[2] = quatPointSetNew[i]->k;
					break;
				
				case SHOWWXY:
					vec.d[0] = quatPointSetNew[i]->r;
					vec.d[1] = quatPointSetNew[i]->i;
					vec.d[2] = quatPointSetNew[i]->j;
					break;

				case SHOWWXZ:
					vec.d[0] = quatPointSetNew[i]->r;
					vec.d[1] = quatPointSetNew[i]->i;
					vec.d[2] = quatPointSetNew[i]->k;
					break;

				default:
					break;
			}
			glVertex3f(vec.d[0], vec.d[1], vec.d[2]);
			//printf("[%d] quat = %f,%f,%f\n",i,vec.d[0],vec.d[1],vec.d[2]);
		}

		glEnd();
	}

	if(showQuatPoints) {

		//glDisable(GL_COLOR_MATERIAL);

		for(i=0;i<numqset;i++) {
			glEnable(GL_LIGHTING);
			glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT,	mat_ambient5);
			glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE,	mat_diffuse5);
			glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR,	mat_specular5);
			glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS,	low_shininess5);
			glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION,	no_mat5);
			glColorMaterial(GL_FRONT_AND_BACK, GL_TRUE);

			if(i==quatkeyframe)
				glEnable(GL_COLOR_MATERIAL);
			else
				glDisable(GL_COLOR_MATERIAL);
			if(i%10==0 || i==(numqset-1) || (i==quatkeyframe) ) {
				switch(whichQuatComponent) {

					case SHOWXYZ:
						vec.d[0] = quatPointSetNew[i]->i;
						vec.d[1] = quatPointSetNew[i]->j;
						vec.d[2] = quatPointSetNew[i]->k;
						break;

					case SHOWWYZ:
						vec.d[0] = quatPointSetNew[i]->r;
						vec.d[1] = quatPointSetNew[i]->j;
						vec.d[2] = quatPointSetNew[i]->k;
						break;
					
					case SHOWWXY:
						vec.d[0] = quatPointSetNew[i]->r;
						vec.d[1] = quatPointSetNew[i]->i;
						vec.d[2] = quatPointSetNew[i]->j;
						break;

					case SHOWWXZ:
						vec.d[0] = quatPointSetNew[i]->r;
						vec.d[1] = quatPointSetNew[i]->i;
						vec.d[2] = quatPointSetNew[i]->k;
						break;

					default:
						break;
				}

				//if(i==quatkeyframe) {
				//	glEnable(GL_COLOR_MATERIAL);
				//	glColor3f(0.1,0.3,0.1);
				//	//glPushMatrix();
				//	//	glTranslatef(vec.d[0], vec.d[1], vec.d[2]);
				//	//	glutSolidSphere(0.043,16,16);
				//	//glPopMatrix();
				//} 
				//else
				//	glDisable(GL_COLOR_MATERIAL);
				//else {
					//if(i==0 ) {
					//	glEnable(GL_COLOR_MATERIAL);
					//	glColor3f(0.1f,0.1f,0.1f);
					//}
					//else {
					//	glDisable(GL_COLOR_MATERIAL);
					//}
				glPushMatrix();
					glTranslatef(vec.d[0], vec.d[1], vec.d[2]);
					glutSolidSphere(0.033,16,16);
				glPopMatrix();
				//}				
			}
		}
	}	
	glPopAttrib();

	glPushAttrib( GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_LINE_BIT );
		glDisable(GL_LIGHTING);
        //tanslucency
        //glEnable(GL_BLEND);
        //glBlendFunc(GL_SRC_COLOR,GL_DST_COLOR);

		glLineWidth(0.01);
		glColor4f(0.8f,0.8f,0.8f,1.0f);
		vec.d[0] = quatPointSetNew[0]->r;
		vec.d[1] = quatPointSetNew[0]->j;
		vec.d[2] = quatPointSetNew[0]->k;
		glPushMatrix();
		glRotatef(90.0f,0.0f,1.0f,0.0f);
		glutWireSphere( length3D(vec.d),20,20 );
        //glutSolidSphere( length3D(vec.d),32,32 );
		glPopMatrix();
        //glDisable(GL_BLEND);
	glPopAttrib();

}



void QuaternionSet::renderThickenedCurve( double radius ) {

	double			p1[3];					//point 1
	double			p2[3];					//point 2
	double			v[3];					//vector p2-p1
	double			r[3];					//vector
	double			theta1	= 0.0;
	double			theta2	= 0.0;
	//int				slices	= 16;
	//int				stacks	= 16;
	double			length	= 0.0;
	//double			zero	= 0.0;
	int				renderPoints = numpset;

	for(int i=0;i<(renderPoints-1);i++) {

		p1[0] = tbset[i].p3.d[0];
		p1[1] = tbset[i].p3.d[1];
		p1[2] = tbset[i].p3.d[2];
		
		p2[0] = tbset[i+1].p3.d[0];
		p2[1] = tbset[i+1].p3.d[1];
		p2[2] = tbset[i+1].p3.d[2];

		minus3D(p2,p1,v);

		r[0] = -v[1]*v[2];
		r[1] =  v[0]*v[2];
		r[2] =  0.0;

		length = sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] );
		double zero = 1.0e-3;

		if (fabs(v[2]) < zero) {
			// rotation angle in x-y plane
			theta1 = (180/M_PI)*acos( v[0] / length );	
			if ( v[0] <= 0.0 ) theta1 = -theta1;
		}
		else {
			// rotation angle
			theta2 = (180/M_PI)*acos( v[2] / length );	
			if ( v[2] <= 0.0 ) theta2 = -theta2;
		}

		glPushAttrib (GL_ALL_ATTRIB_BITS);

			glDisable(GL_CULL_FACE);
			glEnable(GL_LIGHTING);
			glEnable(GL_COLOR_MATERIAL);
			//glColor3f(fgcolor[0],fgcolor[1],fgcolor[2]);
			glColor3f(0.4f,0.7f,0.1f);

			glPushMatrix();
				glTranslatef(p1[0], p1[1], p1[2] ); 
				gluSphere (qobjSPH, radius-0.001, 16, 16);
			glPopMatrix();
			if(i==(renderPoints-2)) {
				glPushMatrix();
					glTranslatef(p2[0], p2[1], p2[2] ); 
					gluQuadricTexture(qobjSPH, GL_TRUE);
					gluSphere (qobjSPH, radius-0.0001, 16, 16);
				glPopMatrix();
			}
			glPushMatrix();
				// translate to point 1
				glTranslatef(p1[0], p1[1], p1[2] );			
    if(fabs(v[2])<=zero) {
     glRotatef(90.0,  0.0, 1.0, 0.0);
     if((v[0]<0.0 && v[1]>0.0) || (v[0]>0.0 && v[1]<0.0))
      glRotatef(theta1, 1.0, 0.0, 0.0);
     else if ((v[0]<0.0 && v[1]<0.0) || (v[0]>0.0 && v[1]>0.0))
      glRotatef(theta1, -1.0, 0.0, 0.0);
    }
    else {
     // Rotate about rotation vector
     glRotatef(theta2, r[0], r[1], r[2]); 
    }
    gluQuadricTexture(qobjCYL, GL_TRUE);
    gluCylinder (qobjCYL, radius, radius, length+0.01, 16, 16);
   glPopMatrix();
  glPopAttrib();

 }
}

void QuaternionSet::printStatus ( )
{
 int i = 0;
 for(i=0; i<numtset; i++) {
  printf("[%d] q = %f, %f, %f %f\n",i,
   quatPointSet[i]->r,quatPointSet[i]->i,
   quatPointSet[i]->j,quatPointSet[i]->k);
 }
}
